#!/usr/bin/env bash

rm -rf \
packages/epanetmsx/toolkit.py \
packages/epanetmsx/*.so \
packages/epanetmsx/*.dll \
packages/epanetmsx/*.dylib \
packages/epanetmsx/*.pyd \
_skbuild \
_cmake_test_compile \
dist \
packages/epanetmsx.egg-info \
